# -*- coding: utf-8 -*-

"""
@author: <sylvain.herledan@oceandatalab.com>
@date: 2017-05-02
"""

import os
import sys
import numpy
import pyproj
import struct
import socket
import logging
import numpy.ma
import datetime
import tempfile
import argparse
import threading
import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
try:
    import cPickle as pickle
except ImportError:
    import pickle

import SEAScope.cmds.get_extraction
import SEAScope.cmds.add_collection
import SEAScope.cmds.add_variable
import SEAScope.cmds.add_granule
import SEAScope.cmds.set_rendering
from SEAScope.upload import connect

logger = logging.getLogger()
handler = logging.StreamHandler()
handler.setLevel(logging.DEBUG)
logger.addHandler(handler)
logger.setLevel(logging.INFO)


def read_response(s, msg_size):
    """"""
    chunk_size = msg_size
    already_read = 0
    result = b''
    while already_read < msg_size:
        buf = s.recv(chunk_size)
        already_read += len(buf)
        result += buf
        chunk_size = msg_size - already_read

    return result


def get_extracted_data():
    """"""
    logger.info('Retrieving extracted data...')
    with connect('localhost', 11155) as s:
        serializer = SEAScope.cmds.get_extraction.serialize
        builder = None
        builder, serialized = serializer(builder)
        builder.Finish(serialized)
        buf = builder.Output()
        s.sendall(struct.pack('>Q', len(buf))+buf)

        buf = s.recv(8)
        l = struct.unpack('<Q', buf)
        logger.debug('Read {} bytes'.format(l[0]))
        buf = read_response(s, l[0])
        logger.debug('{} bytes read'.format(len(buf)))
        deserializer = SEAScope.cmds.get_extraction.deserialize
        raw_result = deserializer(buf)

    clean_data = {}
    for granule in raw_result:
        uri = granule['uri'].decode('utf-8')
        clean_data[uri] = []

        shape = (granule['yArity'], granule['xArity'])
        logger.debug(shape)
        logger.debug(granule['channels'])
        logger.debug(granule['fill_values'])
        logger.debug('# GCPs: {}'.format(len(granule['gcps'])))

        d = numpy.array(granule['buffer'], dtype=numpy.float32)
        logger.debug('Full shape: {}'.format(d.shape))
        fill_values = granule['fill_values']

        d0 = d[::4]
        data = [d0]

        if 1 < len(fill_values):
            d1 = d[1::4]
            data.append(d1)

        if 2 < len(fill_values):
            d2 = d[2::4]
            data.append(d2)

        for n in range(len(data)):
            dn_flipped = numpy.flipud(numpy.reshape(data[n], shape))
            dn_masked = numpy.ma.masked_values(dn_flipped, fill_values[n])
            dn_clean = numpy.ma.masked_invalid(dn_masked)
            clean_data[uri].append(dn_clean)
    logger.info('Done.')
    return clean_data, raw_result


def plot_2d_data(granule_uri, granule_data, output_dir):
    """"""
    fig = plt.figure(1, figsize=(10, 10), dpi=300)
    uri = os.path.basename(granule_uri)
    wrapped_uri = [uri[i:i+50] for i in range(0, len(uri), 50)]
    plt.axis('off')
    plt.title('\n'.join(wrapped_uri), fontsize=8)
    plt.imshow(granule_data[0], cmap=plt.get_cmap('jet'),
    interpolation='nearest')
    output_file = '{}.png'.format(uri)
    output_path = os.path.join(output_dir, output_file)
    fig.savefig(output_path, bbox_inches='tight')
    plt.close()


def plot_2x2d_data(granule_uri, granule_data, output_dir):
    """"""
    d0 = numpy.square(granule_data[0])
    d1 = numpy.square(granule_data[1])
    to_plot = numpy.sqrt(d0 + d1)

    fig = plt.figure(1, figsize=(10, 10), dpi=300)
    uri = os.path.basename(granule_uri)
    wrapped_uri = [uri[i:i+50] for i in range(0, len(uri), 50)]
    plt.axis('off')
    plt.title('\n'.join(wrapped_uri), fontsize=8)
    plt.imshow(to_plot, cmap=plt.get_cmap('jet'), interpolation='nearest')
    output_file = '{}.png'.format(uri)
    output_path = os.path.join(output_dir, output_file)
    fig.savefig(output_path, bbox_inches='tight')
    plt.close()


def plot_rgb_data(granule_uri, granule_data, output_dir):
    """"""
    ubyte_data = []
    for n in range(0,3):
        d_min = numpy.min(granule_data[n])
        d_max = numpy.max(granule_data[n])
        d_ubyte = 255 * (granule_data[n] - d_min) / (d_max - d_min)
        ubyte_data.append(d_ubyte)
    to_plot = numpy.stack(ubyte_data, axis=2).astype(numpy.ubyte)

    fig = plt.figure(1, figsize=(10, 10), dpi=300)
    uri = os.path.basename(granule_uri)
    wrapped_uri = [uri[i:i+50] for i in range(0, len(uri), 50)]
    plt.axis('off')
    plt.title('\n'.join(wrapped_uri), fontsize=8)
    plt.imshow(to_plot)
    output_file = '{}.png'.format(uri)
    output_path = os.path.join(output_dir, output_file)
    fig.savefig(output_path, bbox_inches='tight')
    plt.close()


def serialize_extraction():
    """"""
    logger.info('Serializeing extracted data...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%h%m%s'))
    output_path = os.path.join(output_dir, 'seascope_extract.pyo')
    granules, raw = get_extracted_data()
    with open(output_path, 'wb') as f:
        pickle.dump(raw, f, protocol=1)
    logger.info('Done.')
    return True, '=> Serialization output saved in {}'.format(output_path)


def plot_extraction():
    """"""
    logger.info('Saving extracted data as PNGs...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%h%m%s'))
    granules, raw = get_extracted_data()

    for granule_uri in granules:
        granule_data = granules[str(granule_uri)]
        l = len(granule_data)
        if 1 == l:
            plot_2d_data(granule_uri, granule_data, output_dir)
        elif 2 == l:
            plot_2x2d_data(granule_uri, granule_data, output_dir)
        elif 3 == l:
            plot_rgb_data(granule_uri, granule_data, output_dir)
        else:
            return False, 'Invalid number of channels'
    logger.info('Done.')
    return True, '=> PNGs generated for all granules in {}'.format(output_dir)


def get_dists(granules, raw, geod):
    """"""
    dists = []
    for k in range(len(raw)):
        raw_granule = raw[k]
        granule_uri = raw_granule['uri']
        if b'mask' != granule_uri:
            continue

        clean_data = granules[granule_uri.decode('utf-8')]
        logger.debug('Mask found')
        logger.debug(clean_data)
        gcps = raw_granule['gcps']
        dists.append(0)
        last_lon = gcps[0]['lon']
        last_lat = gcps[0]['lat']
        logger.debug('Number of steps: {}'.format(len(gcps)))
        for n in range(1, len(gcps)):
            lon = gcps[n]['lon']
            lat = gcps[n]['lat']
            _, _, d = geod.inv(last_lon, last_lat, lon, lat)
            print('{} {} -> {} {} = {}'.format(last_lon, last_lat, lon, lat, d))
            dists.append(dists[-1] + .001 * d)
            last_lon = lon
            last_lat = lat
        break
    return dists


def plot_transect():
    """"""
    logger.info('Saving extracted data as PNGs...')
    now = datetime.datetime.utcnow()
    output_dir = tempfile.mkdtemp(suffix=now.strftime('%Y%m%dT%h%m%s'))
    granules, raw = get_extracted_data()

    fig = plt.figure(1, figsize=(5, 5), dpi=300)
    cmap = plt.cm.get_cmap('hsv', 1+len(raw))
    geod = pyproj.Geod(ellps='WGS84')
    dists = get_dists(granules, raw, geod)
    for k in range(len(raw)):
        raw_granule = raw[k]
        granule_uri = raw_granule['uri']
        if b'mask' == granule_uri:
            continue
        clean_data = granules[granule_uri.decode('utf-8')]
        logger.debug(granule_uri)
        logger.debug(clean_data)
        #fig = plt.figure(1, figsize=(5, 5), dpi=300)
        uri = os.path.basename(granule_uri)
        wrapped_uri = [uri[i:i+50] for i in range(0, len(uri), 50)]
        #plt.title('\n'.join(wrapped_uri), fontsize=8)
        plt.xlabel('Distance (km)')
        if 1 == len(clean_data):
            plt.plot(dists, clean_data[0][0, :], c=cmap(k), label=uri)
        elif 2 == len(clean_data):
            values = numpy.sqrt(numpy.square(clean_data[0][0]) + numpy.square(clean_data[1][0]))
            plt.plot(dists, values,c=cmap(k), label=uri)
        #output_file = '{}.png'.format(uri)
        #output_path = os.path.join(output_dir, output_file)
        #fig.savefig(output_path, bbox_inches='tight')
        #plt.close()
    legend = plt.legend(bbox_to_anchor=(0., 1.02, 1., .102), loc=3,
               ncol=1, mode="expand", borderaxespad=0.)
    plt.setp(legend.get_title(),fontsize='xx-small')

    output_file = '{}.png'.format('transects')
    output_path = os.path.join(output_dir, output_file)
    fig.savefig(output_path, bbox_inches='tight')
    plt.close()
    logger.info('Done.')
    return True, '=> Result saved in {}'.format(output_path)


class ProcessingUnit(threading.Thread):
    """"""

    def __init__(self, s):
        """"""
        super(ProcessingUnit, self).__init__()
        self.socket = s

    def run(self):
        """"""
        buf = self.socket.recv(8)
        _opcode = struct.unpack('<Q', buf)
        opcode = _opcode[0]
        logger.debug('Received opcode {}'.format(opcode))
        if 0 == opcode:
            # noop
            pass
        elif 1 == opcode:
            ok, msg = plot_extraction()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        elif 2 == opcode:
            ok, msg = plot_transect()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        elif 3 == opcode:
            ok, msg = serialize_extraction()
            if not ok:
                logger.error(msg)
                return
            logger.info(msg)
        else:
            logger.error('Received unknown opcode {}'.format(opcode))

    def close(self):
        """"""
        self.socket.close()


def serve_forever():
    """"""
    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(('127.0.0.1', 53450))
        s.listen(15)
    except socket.error:
        return False, 'Failed to create socket'

    logger.info('Processing server is running.')
    while True:
        try:
            client_socket, address = s.accept()
            processing_unit = ProcessingUnit(client_socket)
            processing_unit.start()
        except KeyboardInterrupt:
            return True, None

    return False, 'Something went wrong'


def seascope_processor():
    """"""
    ok, msg = serve_forever()
    if not ok:
        logger.error(msg)
        sys.exit(1)
